package de.dfki.madm.mlwizard.cli;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.nio.charset.Charset;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.LogManager;

import org.w3c.dom.Document;

import com.rapidminer.Process;
import com.rapidminer.RapidMiner;
import com.rapidminer.RapidMiner.ExecutionMode;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.io.process.XMLImporter;
import com.rapidminer.io.process.XMLTools;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.io.XrffExampleSetWriter;
import com.rapidminer.operator.io.XrffExampleSource;
import com.rapidminer.operator.meta.ParameterSet;
import com.rapidminer.operator.meta.ParameterValue;
import com.rapidminer.tools.LogService;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.ParameterService;
import com.rapidminer.tools.XMLException;
import com.rapidminer.tools.container.Pair;

import de.dfki.madm.mlwizard.Classifier;
import de.dfki.madm.mlwizard.KnowledgeBase;
import de.dfki.madm.mlwizard.Util;
import de.dfki.madm.mlwizard.functionality.Evaluator;
import de.dfki.madm.mlwizard.functionality.Regressioner;
import de.dfki.madm.mlwizard.functionality.Regressioner.RegressionResult;
import de.dfki.madm.mlwizard.functionality.SystemConstructor;
import de.dfki.madm.mlwizard.metafeatures.MetaFeaturesOperator;

public class CommandLineInterface {

    protected static void printUsage() {
        System.out.println("Arguemnts: <dataset> <task> [outfile]");
        System.out.println("    <dataset>   :   path of the dataset in XRFF format");
        System.out.println("       <task>   :   name of the task to run:");
        System.out.println("                       wizard       - runs the complete wizard");
        System.out.println("                       metafeatures - computes the meta-features");
        System.out.println("                       recommend    - recommends classifers");
        System.out.println("                       evaluate     - evaluates all classifiers");
        System.out.println("                       construct    - constructs the classification system");
        System.out.println("    [outfile]   :   name of the file for the results (default: stdout)");
    }
    
    public static ExampleSet read(String filename) throws OperatorCreationException, OperatorException, IOException {
        XrffExampleSource reader = OperatorService.createOperator(XrffExampleSource.class);
        reader.setParameter(XrffExampleSource.PARAMETER_DATA_FILE, filename);
        return reader.read();
    }
    
    public static void metafeatures(ExampleSet dataset, OutputStream outStream) throws OperatorCreationException, OperatorException, IOException, XMLException {
        MetaFeaturesOperator metaFeaturesOperator = OperatorService.createOperator(MetaFeaturesOperator.class);
        ExampleSet metaFeatures = metaFeaturesOperator.apply(dataset);
        XrffExampleSetWriter.writeXrff(metaFeatures, outStream, Charset.forName("UTF-8")); 
    }
    
    public static RegressionResult recommend(ExampleSet dataset, KnowledgeBase knowledgeBase, PrintStream outStream) throws OperatorCreationException, OperatorException, IOException, XMLException {
        RegressionResult result = Regressioner.predict(knowledgeBase, dataset);
        Collections.sort(result.predictions, new Comparator<Pair<Classifier, Double>>() {
            @Override
            public int compare(Pair<Classifier, Double> arg0, Pair<Classifier, Double> arg1) {
                return arg1.getSecond().compareTo(arg0.getSecond());
            }
        });
        
        int i=0;
        for(Pair<Classifier, Double> prediction:result.predictions) {
            outStream.println(String.format("[%2d] %-18s %.2f", i++, prediction.getFirst().getName(), prediction.getSecond()));
        }
        return result;
    }
    
    private static void print(int i, String classifierName, ParameterSet params, PrintStream outStream) {
        outStream.println(String.format("[%2d] %-18s %.2f", i++, classifierName, params.getPerformance().getMainCriterion().getAverage()));
        for(Iterator<ParameterValue> it = params.getParameterValues(); it.hasNext();) {
            ParameterValue p = it.next();
            outStream.println(String.format("          %-28s %s", (p.getOperator() + "." + p.getParameterKey()), p.getParameterValue()));
        }
    }
        
    public static Map<Classifier, ParameterSet> optimize(ExampleSet dataset, Example queryMetaFeatures, Collection<Classifier> classifiers, int nThreads, KnowledgeBase knowledgeBase, PrintStream outStream, RegressionResult regressionResult) throws OperatorCreationException, OperatorException, IOException, XMLException {
        Evaluator evaluator = new Evaluator();
        Map<Classifier, ParameterSet> result = evaluator.evaluate(classifiers, queryMetaFeatures, knowledgeBase, dataset, nThreads, null);
        try {
            evaluator.awaitTermination();
            int i = 0;
            if(regressionResult == null) {
                for(Classifier classifier:classifiers) {
                    print(i++, classifier.getName(), result.get(classifier), outStream);
                }
                for(Classifier classifier:knowledgeBase.getClassifiers()) {
                    if(!classifiers.contains(classifier)) {
                        outStream.println(String.format("[%2d] %-18s <not evaluated>", i++, classifier.getName()));
                    }
                }
            } else {
                for(Pair<Classifier, Double> prediction:regressionResult.predictions) {
                    Classifier classifier = prediction.getFirst();
                    if(result.containsKey(classifier)) {
                        print(i++, classifier.getName(), result.get(classifier), outStream);
                    } else {
                        outStream.println(String.format("[%2d] %-18s <not evaluated>", i++, classifier.getName()));
                    }
                }
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        return result;
    }
    
    public static void construct(ExampleSet dataset, Classifier classifier, ParameterSet parameterSet, String filename, OutputStream outStream) throws OperatorCreationException, OperatorException, IOException, XMLException {
        Process process = SystemConstructor.createProcess(classifier, parameterSet, filename, true);
        Document document = process.getRootOperator().getDOMRepresentation();
        XMLTools.stream(document, outStream, XMLImporter.PROCESS_FILE_CHARSET);
    }
    
    public static void wizard(ExampleSet dataset, KnowledgeBase knowledgeBase, String filename, OutputStream outStream) throws OperatorCreationException, OperatorException, IOException, XMLException {
        RegressionResult result = recommend(dataset, knowledgeBase, System.out);
        Vector<Classifier> classifiers = new Vector<Classifier>();
        System.out.println("Enter comma separated numbers of classifiers you want to be evaluated:");
        BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
        String line;
        while((line = reader.readLine().trim()).isEmpty());
        HashSet<Integer> selected = new HashSet<Integer>();
        for(String s:line.split(",")) {
            try {
                selected.add(Integer.parseInt(s));
            } catch(NumberFormatException e) {
                System.err.println("No number: " + s.trim());
            }
        }
        for(int i=0; i<result.predictions.size(); i++) {
            if(selected.contains(i)) {
                classifiers.add(result.predictions.get(i).getFirst());
            }
        }
        Map<Classifier, ParameterSet> result2 = optimize(dataset, result.metaFeatures.getExample(0), classifiers, Runtime.getRuntime().availableProcessors(), knowledgeBase, System.out, result);
        System.out.println("Enter number of classifier you want a system for:");
        while((line = reader.readLine().trim()).isEmpty());
        try {
            int n = Integer.parseInt(line);
            Classifier classifier = result.predictions.get(n).getFirst();
            construct(dataset, classifier, result2.get(classifier), filename, outStream);
        } catch(NumberFormatException e) {
            System.err.println("No number: " + line);
        }
    }
    

    public static void main(String[] argv) throws OperatorCreationException, OperatorException, IOException, XMLException {
        if(argv.length<2) {
            printUsage();
            System.exit(-1);
        }
        
        LogManager logManager = LogManager.getLogManager();
        logManager.reset();
        ParameterService.setParameterValue(RapidMiner.PROPERTY_RAPIDMINER_GLOBAL_LOG_VERBOSITY, String.valueOf(LogService.ERROR));
        LogService.getGlobal().setVerbosityLevel(LogService.ERROR);
        LogService.getRoot().setLevel(Level.SEVERE);
        RapidMiner.setExecutionMode(ExecutionMode.COMMAND_LINE);
        RapidMiner.init();
        
        String filename = argv[0];
        String task = argv[1].toLowerCase().trim();
        PrintStream outStream = System.out;
        if(argv.length>2) {
            outStream = new PrintStream(argv[2]);
        }
        
        KnowledgeBase knowledgeBase = KnowledgeBase.read();
        ExampleSet dataset = read(filename);
        
        if(task.equals("wizard")) {
            wizard(dataset, knowledgeBase, filename, outStream);
        } else if(task.equals("metafeatures")) {
            metafeatures(dataset, outStream);
        } else if(task.equals("recommend")) {
            recommend(dataset, knowledgeBase, outStream);
        } else if(task.equals("evaluate")) {
            MetaFeaturesOperator metaFeaturesOperator = OperatorService.createOperator(MetaFeaturesOperator.class);
            Example metaFeatures = metaFeaturesOperator.apply(dataset).getExample(0);
            optimize(dataset, metaFeatures, knowledgeBase.getClassifiers(), Runtime.getRuntime().availableProcessors(), knowledgeBase, outStream, null);
        } else if(task.equals("construct")) {
            List<String[]> bounds = new LinkedList<String[]>();
            Classifier classifier = new Classifier("complex", Util.readFileAsString("/tmp/preprocess.xml"), Util.readFileAsString("/tmp/learn.xml"), bounds, bounds);
            construct(dataset, classifier, null, filename, outStream);
        } else {
            System.err.println("unknown task: " + task);
            printUsage();
        }
        outStream.close();
    }
}
