package de.dfki.madm.mlwizard.functionality;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;

import com.rapidminer.Process;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.IOContainer;
import com.rapidminer.operator.MissingIOObjectException;
import com.rapidminer.operator.OperatorCapability;
import com.rapidminer.operator.OperatorChain;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.ProcessStoppedException;
import com.rapidminer.operator.learner.AbstractLearner;
import com.rapidminer.operator.meta.OperatorSelector;
import com.rapidminer.operator.meta.ParameterSet;
import com.rapidminer.operator.performance.PerformanceVector;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.tools.AttributeSubsetSelector;
import com.rapidminer.parameter.ParameterTypeList;
import com.rapidminer.tools.XMLException;

import de.dfki.madm.mlwizard.Classifier;
import de.dfki.madm.mlwizard.KnowledgeBase;
import de.dfki.madm.mlwizard.Util;
import de.dfki.madm.mlwizard.optimization.ESOptimizationPaREn;

/**
 * @brief evaluates a set of classifiers on a given dataset
 */
public class Evaluator {

    /** the actual dataset */
    private ExampleSet dataset;
    /** the evaluation results */
    private Map<Classifier, ParameterSet> results;

    private ExecutorService threads;
    private HashMap<String, EvaluateSwingThread> workers;
    private EvaluationListener evaluationListener;

    /**
     * creates a new evaluator
     */
    public Evaluator() {
        results = Collections.synchronizedMap(new HashMap<Classifier, ParameterSet>());
        workers = new HashMap<String, EvaluateSwingThread>();
    }

    public Map<Classifier, ParameterSet> evaluate(Iterable<Classifier> classifiers,
            Example queryMetaFeatures,
            KnowledgeBase knowledgeBase,
            ExampleSet dataset,
            int nThreads) throws OperatorCreationException, IOException, XMLException {
        return evaluate(classifiers, queryMetaFeatures, knowledgeBase, dataset, nThreads, null);
    }

    
    /**
     * evaluates a set of classifier on one dataset
     * @param classifierNames the set of classifier names
     * @param queryMetaFeatures the meta-features of the dataset
     * @param knowledgeBase the knowledge base to use
     * @param dataset the actual dataset
     * @param nThreads number of threads for the evaluation
     * @param evaluationListener the listener of evaluation events (might be null)
     * @return a map of classifiers and their optimized parameter values
     * @throws OperatorCreationException
     * @throws IOException
     * @throws XMLException
     */
    public Map<Classifier, ParameterSet> evaluate(Iterable<Classifier> classifiers,
                                              Example queryMetaFeatures,
                                              KnowledgeBase knowledgeBase,
                                              ExampleSet dataset,
                                              int nThreads,
                                              EvaluationListener evaluationListener) throws OperatorCreationException, IOException, XMLException {
        this.dataset = dataset;
        this.evaluationListener = evaluationListener;
        if(nThreads <= 0) {
            nThreads = Runtime.getRuntime().availableProcessors();
        }
        threads = Executors.newFixedThreadPool(nThreads);
        results.clear();
        workers.clear();
        for(Classifier classifier:classifiers) {
            int populationSize = 7;
            int stop = 2;
            
            Process process = new Process(Util.getStream("parameter_optimization.xml"));
            String optimizationCase = String.valueOf(classifier.getOptimizationCase());
            OperatorChain validation = (OperatorChain)process.getOperator("Validation " + optimizationCase);
            
            // -- set optimization parameters --
            process.getOperator("Select Optimization").setParameter(OperatorSelector.PARAMETER_SELECT_WHICH, optimizationCase);
            if(!classifier.getNominalParmas().isEmpty()) {
                process.getOperator("Grid " + optimizationCase).setParameter("parameters", ParameterTypeList.transformList2String(classifier.getNominalParmas()));
            }
            // -- if classifier has numerical parameters, we use a GA with defined start points --
            if(!classifier.getNumericalParams().isEmpty()) {
                ESOptimizationPaREn optimizer = (ESOptimizationPaREn)process.getOperator("ESOptimizationPaREn " + optimizationCase);
                optimizer.setParameter("parameters", ParameterTypeList.transformList2String(classifier.getNumericalParams()));
                optimizer.setParameter(ESOptimizationPaREn.PARAMETER_POPULATION_SIZE, String.valueOf(populationSize));
                optimizer.setParameter(ESOptimizationPaREn.PARAMETER_GENERATIONS_WITHOUT_IMPROVAL, String.valueOf(stop));
                optimizer.setStartValues(knowledgeBase.getStartPoints(classifier, queryMetaFeatures, populationSize));
            }
            
            // -- fixed pre-processing with single-operator learner --
            if(classifier.getLearner() != null) {
                // -- enable fixed pre-processing --
                process.getOperator("Select Preprocess").setParameter(OperatorSelector.PARAMETER_SELECT_WHICH, "1");
                
                AbstractLearner learner = classifier.getLeanerInstance();
                // -- (de)activate nominal to numerical conversation -- 
                boolean supportsNominal = learner.supportsCapability(OperatorCapability.BINOMINAL_ATTRIBUTES) && learner.supportsCapability(OperatorCapability.POLYNOMINAL_ATTRIBUTES);
                process.getOperator("Nominal to Binominal").setParameter(AttributeSubsetSelector.PARAMETER_INVERT_SELECTION, String.valueOf(supportsNominal));
                process.getOperator("Nominal to Numerical").setParameter(AttributeSubsetSelector.PARAMETER_INVERT_SELECTION, String.valueOf(supportsNominal));
                
                // -- inject learner --
                validation.getSubprocess(0).addOperator(learner);
                validation.getSubprocess(0).getInnerSources().getPortByIndex(0).connectTo(learner.getInputPorts().getPortByIndex(0));
                learner.getOutputPorts().getPortByIndex(0).connectTo(validation.getSubprocess(0).getInnerSinks().getPortByIndex(0));
                
            // -- custom pre-process and learning process --
            } else {
                // -- enable custom pre-processing --
                process.getOperator("Select Preprocess").setParameter(OperatorSelector.PARAMETER_SELECT_WHICH, "2");
                
                // -- inject preprocess --
                OperatorChain preProcessChain = (OperatorChain)process.getOperator("Select Preprocess");
                Process p = new Process(classifier.getPreProcess());
                InputPort input = p.getRootOperator().getSubprocess(0).getInnerSources().getPortByIndex(0).getDestination();
                OutputPort output = p.getRootOperator().getSubprocess(0).getInnerSinks().getPortByIndex(0).getSource();
                preProcessChain.getSubprocess(1).stealOperatorsFrom(p.getRootOperator().getSubprocess(0));
                preProcessChain.getSubprocess(1).getInnerSources().getPortByIndex(0).connectTo(input);
                output.connectTo(preProcessChain.getSubprocess(1).getInnerSinks().getPortByIndex(0));
                
                // -- inject learning process --
                Process p2 = new Process(classifier.getLearnProcess());
                input = p2.getRootOperator().getSubprocess(0).getInnerSources().getPortByIndex(0).getDestination();
                output = p2.getRootOperator().getSubprocess(0).getInnerSinks().getPortByIndex(0).getSource();
                validation.getSubprocess(0).stealOperatorsFrom(p2.getRootOperator().getSubprocess(0));
                validation.getSubprocess(0).getInnerSources().getPortByIndex(0).connectTo(input);
                output.connectTo(validation.getSubprocess(0).getInnerSinks().getPortByIndex(0));
            }
            process.save(new File("/tmp/"+classifier.getName()+".rmp"));
            
            EvaluateSwingThread evaluateThread = new EvaluateSwingThread(classifier, process);
            workers.put(classifier.getName(), evaluateThread);
            threads.execute(evaluateThread);
        }
        threads.shutdown();
        return results;
    }
    
    public void awaitTermination() throws InterruptedException {
        threads.awaitTermination(1, TimeUnit.DAYS);
    }

    /**
     * cancels the evaluation thread of a given classifier
     * @param classifierName the name of the classifier which evaluation should be stopped
     */
    public void cancelThread(String classifierName) {
        workers.get(classifierName).stop();
        if (!workers.get(classifierName).isDone()) {
            workers.get(classifierName).cancel(true);
        }
    }

    /**
     * cancel all evaluation threads
     */
    public void cancelThreads() {
        for (EvaluateSwingThread evaluateThread : workers.values()) {
            evaluateThread.stop();
        }
        for (EvaluateSwingThread evaluateThread : workers.values()) {
            if (!evaluateThread.isDone()) {
                evaluateThread.cancel(true);
            }
        }
    }

    /**
     * @brief thread for the evaluation/optimization of one classifier
     */
    public class EvaluateSwingThread extends SwingWorker<IOContainer, Object> {
        private final Classifier classifier;
        private Process process;

        public EvaluateSwingThread(final Classifier classifier, Process process) {
            this.classifier = classifier;
            this.process = process;
        }

        public void stop() {
            process.stop();
        }

        @Override
        public IOContainer doInBackground() throws OperatorException {
            try {
                SwingUtilities.invokeLater(new Runnable() {
                    @Override
                    public void run() {
                        if (evaluationListener != null) {
                            evaluationListener.setComputing(classifier.getName());
                        }
                    }
                });
                ExampleSet dataSetCopy = null;
                synchronized (dataset) {
                    dataSetCopy = (ExampleSet) dataset.clone();
                }
                IOContainer result = process.run(new IOContainer(dataSetCopy));

                PerformanceVector performanceVector = result.get(PerformanceVector.class);
                ParameterSet p1 = null;
                ParameterSet p2 = null;
                try {
                    p1 = result.get(ParameterSet.class, 0);
                    p2 = result.get(ParameterSet.class, 1);
                } catch (MissingIOObjectException e) {
                    // -- no actual error --
                }
                ParameterSet parameterSet = Util.combineParameterSets(p1, p2, performanceVector);
                if (evaluationListener != null) {
                    synchronized (evaluationListener) {
                        evaluationListener.setPerformance(classifier.getName(), performanceVector);
                    }
                }
                results.put(classifier, parameterSet);
                return result;
            } catch (ProcessStoppedException e1) {
            }
            return null;
        }

//        @Override
//        protected void done() {
//            System.out.println("done");
//            if (!isCancelled()) {
//                try {
//                    IOContainer result = get();
//                } catch (Exception e) {
//
//                }
//            }
//        }

    }

    /**
     * @brief interface of listeners for evaluation events
     */
    public interface EvaluationListener {
        /**
         * a classifier is now being evaluated
         * @param classifierName the name of the classifier
         */
        public void setComputing(String classifierName);
        /**
         * the evaluation of a classifier is finished
         * @param classifierName the name of the classifier
         * @param performanceVector the achieved evaluation result
         */
        public void setPerformance(String classifierName, PerformanceVector performanceVector);
    }

}
