/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.mlwizard.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.image.BufferedImage;
import java.util.HashMap;

import javax.swing.AbstractButton;
import javax.swing.ButtonGroup;
import javax.swing.DefaultCellEditor;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JProgressBar;
import javax.swing.JRadioButton;
import javax.swing.JTable;
import javax.swing.event.ChangeListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

import com.rapidminer.example.Attribute;
import com.rapidminer.gui.tools.SwingTools;
import com.rapidminer.operator.performance.PerformanceVector;
import com.rapidminer.tools.I18N;
import com.rapidminer.tools.Tools;
import com.rapidminer.tools.container.Pair;

import de.dfki.madm.mlwizard.Classifier;
import de.dfki.madm.mlwizard.functionality.Evaluator;
import de.dfki.madm.mlwizard.functionality.Evaluator.EvaluationListener;
import de.dfki.madm.mlwizard.functionality.Regressioner.RegressionResult;

/**
 * 
 * @author Matthias Reif
 *
 */
public class EvaluateTableModel extends AbstractTableModel implements ActionListener, EvaluationListener {

	private static final long serialVersionUID = 8742646362200822187L;
	private String[] columnNames = { "Classifier", "Predicted Accuracy", "Root Mean Squared Error", "Actual Accuracy", "Cancel", "Load"};
    Attribute classifierAttribute, predictionAttribute, rmseAttribute;
    private HashMap<String, Object> result;
    private HashMap<String, JTableRadioButton> loadButtons;
    private HashMap<String, JTableButton> cancelButtons;
    private HashMap<JTableButton, String> cancelButtonsReverse;
    private Evaluator evaluator;
    private RegressionResult regressionResult;
    
    private static final int CANCEL_COLUMN = 4;
    private static final int LOAD_COLUMN = 5;

    public EvaluateTableModel(RegressionResult regressionResult, JTable table, HashMap<String, Boolean> evaluate, Evaluator evaluator) {
        this.regressionResult = regressionResult;
        this.evaluator = evaluator;
        result = new HashMap<String, Object>();
        loadButtons = new HashMap<String, JTableRadioButton>();
        cancelButtons = new HashMap<String, JTableButton>();
        cancelButtonsReverse = new HashMap<JTableButton, String>();
        ButtonGroup buttonGroup = new ButtonGroup();
        int row = 0;
        for(Pair<Classifier, Double> entry:regressionResult.predictions) {
            String classifierName = entry.getFirst().getName();
            JTableRadioButton radioButton = new JTableRadioButton(table, LOAD_COLUMN, row);
            radioButton.setBackground(Color.WHITE);
            buttonGroup.add(radioButton);
            loadButtons.put(classifierName, radioButton);
            JTableButton cancelButton = new JTableButton(table, CANCEL_COLUMN, row);
            cancelButton.addActionListener(this);
            cancelButtons.put(classifierName, cancelButton);
            cancelButtonsReverse.put(cancelButton, classifierName);
            if(evaluate.get(classifierName)) {
                result.put(classifierName, new JTableProgressBar(table, 3, row));
            } else {
                result.put(classifierName, "-");
                cancelButtons.get(classifierName).setEnabled(false);
            }
            row++;
        }
    }

    public void addChangeListener(ChangeListener listener) {
        for(JRadioButton radioButton : loadButtons.values()) {
            radioButton.addChangeListener(listener);
        }
    }

    public boolean isOneSelected() {
        for(String id:loadButtons.keySet()) {
            if(loadButtons.get(id).isSelected()) {
                return true;
            }
        }
        return false;
    }

    public String getSelectedClassifierName() {
        for(String id:loadButtons.keySet()) {
            if(loadButtons.get(id).isSelected()) {
                return id;
            }
        }
        return null;
    }
    
    public void setComputing(String classifierName) {
        Object o = result.get(classifierName);
        if(o instanceof JTableProgressBar) {
            ((JTableProgressBar)o).setIndeterminate(true);
        }
    }
    
    public void setPerformance(String classifierName, PerformanceVector performanceVector) {
        result.put(classifierName, Tools.formatNumber(performanceVector.getMainCriterion().getAverage()));
        cancelButtons.get(classifierName).setEnabled(false);
        fireTableDataChanged();
    }

    @Override
    public int getColumnCount() {
        return columnNames.length;
    }

    @Override
    public int getRowCount() {
        return regressionResult.predictions.size();
    }

    @Override
    public String getColumnName(int col) {
        return columnNames[col];
    }

    @Override
    public Object getValueAt(int row, int col) {
        Pair<Classifier, Double> prediction = regressionResult.predictions.get(row);
        String classifierName = prediction.getFirst().getName();
        switch(col) {
        case 0:
            return classifierName;
        case 1:
            // -- Values >1 may happen because of regression modeling, but looks strange to the user --
            return Tools.formatNumber(Math.min(prediction.getSecond(), 1.0));
        case 2:
            return Tools.formatNumber(prediction.getFirst().getRMSE());
        case 3:
            return result.get(classifierName);
        case CANCEL_COLUMN:
            return cancelButtons.get(classifierName);
        case LOAD_COLUMN:
            return loadButtons.get(classifierName);
        }
        return null;
    }

    public Class <? extends Object> getColumnClass(int col) {
        if(col >= 3) {
            return Object.class;
        } else {
            return String.class;
        }
    }

    public boolean isCellEditable(int row, int col) {
        return col == CANCEL_COLUMN || col == LOAD_COLUMN;
    }
    
    /**
     * cancel on one classifier evaluation clicked
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        String classifierName = cancelButtonsReverse.get(e.getSource());
        evaluator.cancelThread(classifierName);
        result.put(classifierName, "cancelled");
        cancelButtons.get(classifierName).setEnabled(false);
        fireTableDataChanged();
    }
    
    class AbstractButtonEditor extends DefaultCellEditor implements ItemListener {
		private static final long serialVersionUID = -8216786630949657785L;
		protected AbstractButton button;

        public AbstractButtonEditor(JCheckBox checkBox) {
            super(checkBox);
        }
        @Override
        public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row, int column) {
            if(value == null)
                return null;
            button = (AbstractButton) value;
            button.addItemListener(this);
            return (Component) value;
        }
        @Override
        public Object getCellEditorValue() {
            button.removeItemListener(this);
            return button;
        }
        @Override
        public void itemStateChanged(ItemEvent e) {
            super.fireEditingStopped();
        }
    }

    class JTableButton extends JButton {
		private static final long serialVersionUID = 7950996735177291390L;
		private JTable table;
        private int column, row;

        public JTableButton(JTable table, int column, int row) {
            super();
            this.table = table;
            this.column = column;
            this.row = row;
            String iconLocation = I18N.getMessageOrNull(I18N.getGUIBundle(), "gui.action.cancel.icon");
            if (iconLocation != null) {
                ImageIcon smallIcon = SwingTools.createIcon("16/"+iconLocation);
                Image img = smallIcon.getImage();
                BufferedImage bi = new BufferedImage(12, 12, BufferedImage.TYPE_INT_ARGB);
                Graphics g = bi.createGraphics();
                g.drawImage(img, 2, 2, 8, 8, null);
                ImageIcon icon = new ImageIcon(bi);
                setIcon(icon);
            }
        }
        @Override
        public void repaint() {
            // -- to trigger the cell repainting --
            if(table != null && table.getModel() != null) {
                ((AbstractTableModel) table.getModel()).fireTableCellUpdated(row, column);
            }
        }
    }
    
    class JTableRadioButton extends JRadioButton {
		private static final long serialVersionUID = -7096238025090732549L;
		private JTable table;
        private int column, row;

        public JTableRadioButton(JTable table, int column, int row) {
            super();
            this.table = table;
            this.column = column;
            this.row = row;
        }
        @Override
        public void repaint() {
            // -- to trigger the cell repainting --
            if(table != null && table.getModel() != null) {
                ((AbstractTableModel) table.getModel()).fireTableCellUpdated(row, column);
            }
        }
    }
    
    class JTableProgressBar extends JProgressBar {
		private static final long serialVersionUID = -2038531846689892164L;
		private JTable table;
        private int column, row;

        public JTableProgressBar(JTable table, int column, int row) {
            super();
            setIndeterminate(false);
            setStringPainted(false);
            this.table = table;
            this.column = column;
            this.row = row;
        }
        @Override
        public boolean isDisplayable() {
            // -- to get the animation always running --
            return true;
        }
        @Override
        public void repaint() {
            // -- to trigger the cell repainting --
            if(table != null && table.getModel() != null) {
                ((AbstractTableModel) table.getModel()).fireTableCellUpdated(row, column);
            }
        }
    }
    
    class JTableComponentRenderer extends DefaultTableCellRenderer {
		private static final long serialVersionUID = -3435697265654945361L;

		@Override
        public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
            if(value instanceof JComponent) {
                return (JComponent) value;
            } else {
                return super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
            }
        }
    }
}
