/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.mlwizard.gui;

import java.awt.event.WindowEvent;
import java.awt.event.WindowFocusListener;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.table.TableColumnModel;

import com.rapidminer.Process;
import com.rapidminer.gui.RapidMinerGUI;
import com.rapidminer.gui.tools.dialogs.wizards.WizardStep;
import com.rapidminer.operator.meta.ParameterSet;

import de.dfki.madm.mlwizard.Classifier;
import de.dfki.madm.mlwizard.functionality.Evaluator;
import de.dfki.madm.mlwizard.functionality.Regressioner.RegressionResult;
import de.dfki.madm.mlwizard.functionality.SystemConstructor;

/**
 * This is the third step in the Automatic System Construction Wizard.
 * The selected classifiers are evaluated by parameter grid search and x-validation.
 * In the end, one selected classifier will be used as basis for the construction of a 
 * classification process.
 * 
 * @author Matthias Reif, Alexander Arimond
 *
 */

class EvaluateWizardStep extends WizardStep implements WindowFocusListener {

    private RegressionWizardStep regressionWizardStep;
    private JPanel mainPanel;
    private AutomaticSystemConstructionWizard wizard;
    private EvaluateTableModel model;
    private Map<Classifier, ParameterSet> results;
    private Evaluator evaluator;
    
    public EvaluateWizardStep(String i18nKey, RegressionWizardStep regressionWizardStep, AutomaticSystemConstructionWizard parent) {
        super(i18nKey);
        this.regressionWizardStep = regressionWizardStep;
        this.wizard = parent;
        mainPanel = new JPanel();
        mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
        evaluator = new Evaluator();
    }

    @Override
    protected boolean canGoBack() {
        return false;
    }

    @Override
    protected boolean canProceed() {
        return model != null && model.isOneSelected();
    }

    @Override
    protected JComponent getComponent() {
        return mainPanel;
    }
    

    @Override
    public boolean performEnteringAction() {
        initDialog();
        try {
            Vector<Classifier> classifiers = new Vector<Classifier>();
            for(String classifierName:regressionWizardStep.getModel().getIfEvaluate()) {
                classifiers.add(wizard.getKnowledgeBase().getClassifier(classifierName));
            }
            results = evaluator.evaluate(classifiers,
                                         regressionWizardStep.getModel().getRegressionResult().metaFeatures.getExample(0),
                                         wizard.getKnowledgeBase(),
                                         wizard.getDataSet(),
                                         regressionWizardStep.getNThreads(),
                                         model);
        } catch (Exception e) {
            wizard.exception(e);
        }
        return true;
    }

    @Override
    public boolean performLeavingAction() {
        String classifierName = model.getSelectedClassifierName();
        if(classifierName == null) {
            return false;
        }
        Classifier classifier = wizard.getKnowledgeBase().getClassifier(classifierName);
        ParameterSet parameterSet = null;
        if(!results.containsKey(classifier)) {
            String msg = "The selected classifier was not optimized. Do you want to create an unoptimized process?";
            if(JOptionPane.showConfirmDialog(wizard, msg, "Confirmation", JOptionPane.YES_NO_OPTION) == JOptionPane.NO_OPTION) {
                return false;
            }
        } else {
            parameterSet = results.get(classifier);
        }

        try {
            Process process = SystemConstructor.createProcess(wizard.getKnowledgeBase().getClassifier(classifierName),
                                                              parameterSet, wizard.getDataSetRepositoryLocation(), false);
            RapidMinerGUI.getMainFrame().addWindowFocusListener(this);
            RapidMinerGUI.getMainFrame().setProcess(process, true);
        } catch(Exception e) {
            wizard.exception(e);
            return false;
        }
        return true;
    }
    
    
    void initDialog() {
        JTable table = new JTable();

        RegressionResult regressionResult = regressionWizardStep.getModel().getRegressionResult();
        HashMap<String, Boolean> evaluate = regressionWizardStep.getModel().getEvaluate();
        model = new EvaluateTableModel(regressionResult, table, evaluate, evaluator);
        model.addChangeListener(wizard);
        table.setModel(model);

        mainPanel.removeAll();
        mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
        mainPanel.add(RegressionWizardStep.createResultPanel(table, wizard.getDataSet(), regressionWizardStep));


        TableColumnModel dstTCM = table.getColumnModel();
        TableColumnModel srcTCM = regressionWizardStep.getTable().getColumnModel();
        for(int i = 0; i < 4; i++) {
            dstTCM.getColumn(i).setPreferredWidth(srcTCM.getColumn(i).getWidth());
        }
        table.getColumn("Actual Accuracy").setCellRenderer(model.new JTableComponentRenderer());
        
        table.getColumn("Load").setCellRenderer(model.new JTableComponentRenderer());
        table.getColumn("Load").setCellEditor(model.new AbstractButtonEditor(new JCheckBox()));
       
        table.getColumn("Cancel").setCellRenderer(model.new JTableComponentRenderer());
        table.getColumn("Cancel").setCellEditor(model.new AbstractButtonEditor(new JCheckBox()));
    }

    
    public void stopThreads() {
        evaluator.cancelThreads();
    }
    
    @Override
    public void windowGainedFocus(WindowEvent arg0) {
        RapidMinerGUI.getMainFrame().getProcessPanel().getProcessRenderer().ARRANGE_OPERATORS_ACTION.actionPerformed(null);
    }

    @Override
    public void windowLostFocus(WindowEvent arg0) {
    }
    
   
}
