/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.mlwizard.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTable;
import javax.swing.RowSorter;
import javax.swing.SortOrder;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingWorker;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.gui.tools.dialogs.wizards.WizardStep;
import com.rapidminer.tools.Tools;

import de.dfki.madm.mlwizard.functionality.Regressioner;
import de.dfki.madm.mlwizard.functionality.Regressioner.RegressionResult;


/**
 * This is the second step of the PaREn Automatic System Construction Wizard.
 * Predictions on accuracies are made for several classifiers based on previous accuracies (casebase) 
 * and meta features (e.g. landmarking).
 * 
 * @author Matthias Reif, Alexander Arimond
 *
 */

class RegressionWizardStep extends WizardStep implements ChangeListener, ActionListener {

    private JPanel mainPanel, progressPanel, resultPanel;
    private AutomaticSystemConstructionWizard parent;
    private JTable table;
    private RegressionTableModel model;
    private JSpinner nThreadsSpinner;
    private JButton selectAllButton;
    
    private static final int N_THREADS_DEFAULT = 1;
    
    public RegressionWizardStep(String i18nKey, AutomaticSystemConstructionWizard parent) {
        super(i18nKey);
        this.parent = parent;
    }
    
    @Override
    protected boolean canGoBack() {
        return false;
    }

    @Override
    protected boolean canProceed() {
        return model != null;
    }

    public int getNThreads() {
        return (Integer) nThreadsSpinner.getValue();
    }
    
    public RegressionTableModel getModel() {
        return model;
    }
    
    public JTable getTable() {
        return table;
    }
    
    @Override
    protected JComponent getComponent() {
        mainPanel = new JPanel();
        mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));

        progressPanel = new JPanel();
        progressPanel.setLayout(new BoxLayout(progressPanel, BoxLayout.PAGE_AXIS));

        JProgressBar progressBar = new JProgressBar();
        progressBar.setIndeterminate(true);

        progressPanel.add(Box.createRigidArea(new Dimension(20, 20)));
        progressPanel.add(new JLabel("Please wait while analyzing the data..."));
        progressPanel.add(Box.createRigidArea(new Dimension(20, 20)));
        progressPanel.add(progressBar);
        progressPanel.add(Box.createVerticalGlue());

        mainPanel.add(progressPanel);

        return mainPanel;
    }

    protected static JScrollPane getTablePanel(JTable table) {
        table.setAutoCreateRowSorter(true);
        ArrayList<RowSorter.SortKey> sortKeys = new ArrayList<RowSorter.SortKey>();
        sortKeys.add(new RowSorter.SortKey(1, SortOrder.DESCENDING));
        sortKeys.add(new RowSorter.SortKey(2, SortOrder.ASCENDING));
        sortKeys.add(new RowSorter.SortKey(0, SortOrder.ASCENDING));
        table.getRowSorter().setSortKeys(sortKeys);
        JScrollPane scrollPane = new JScrollPane(table);
        scrollPane.setBorder(BorderFactory.createLineBorder(Color.BLACK));
        table.setFillsViewportHeight(true);
        return scrollPane;
    }

    protected void createMainPanel(JComponent mainPanel, RegressionResult result) {
        model = new RegressionTableModel(result);
        table = new JTable(model);
        progressPanel.setVisible(false);
        mainPanel.remove(progressPanel);
        parent.validate();
        resultPanel = createResultPanel(table, parent.getDataSet(), this);

        SpinnerNumberModel spinnerModel = new SpinnerNumberModel(N_THREADS_DEFAULT, 1, Integer.MAX_VALUE, 1);
        
        selectAllButton = new JButton(new AbstractAction("Select All") {
			private static final long serialVersionUID = -5430858335384041735L;

			@Override
            public void actionPerformed(ActionEvent e) {
                for(int i=0; i<model.getRowCount(); i++) {
                    model.setValueAt(true, i, 3);
                }
            }
        });
        Box buttonBox = new Box(BoxLayout.LINE_AXIS);
        buttonBox.add(Box.createHorizontalGlue());
        buttonBox.add(selectAllButton);
        buttonBox.add(Box.createHorizontalGlue());
        resultPanel.add(buttonBox, BorderLayout.PAGE_END);
        
        nThreadsSpinner = new JSpinner(spinnerModel);

        mainPanel.add(resultPanel);
        mainPanel.revalidate();
        parent.validate();
    }

    protected static JPanel createResultPanel(JTable table, ExampleSet dataSet, ActionListener actionListener) {
        String msg = "Dataset has " + dataSet.size() + " samples with " + dataSet.getAttributes().size() + " attributes.";
        Box topBox = Box.createHorizontalBox();
        topBox.add(new JLabel(msg));
        if(actionListener != null) {
            topBox.add(Box.createHorizontalStrut(20));
            ((JButton) topBox.add(new JButton("Show Meta Features"))).addActionListener(actionListener);
        }
        JPanel panel = new JPanel(new BorderLayout(0, 10));
        panel.add(topBox, BorderLayout.PAGE_START);
        panel.add(getTablePanel(table), BorderLayout.CENTER);
        return panel;
    }
    
    @Override
    protected boolean performEnteringAction() {
        new SwingWorker<RegressionResult, Object>() {

            @Override
            public RegressionResult doInBackground() {
                try {
                    return Regressioner.predict(parent.getKnowledgeBase(), parent.readDataSet());
                } catch(Exception e) {
                    parent.exception(e);
                    return null;
                }
            }
            
			@Override
            protected void done() {
                try {
                    createMainPanel(mainPanel, get());
                    parent.stateChanged(new ChangeEvent(this));
                } catch(Exception e) {
                    e.printStackTrace();
                }
            }
        }.execute();
        return true;
    }

    @Override
    public void stateChanged(ChangeEvent arg0) {
    }

    public void showFeatureSet() {
        ExampleSet metaFeatures = model.getRegressionResult().metaFeatures;
        Object[][] values = new Object[metaFeatures.getAttributes().size()][2];
        for(Example example : metaFeatures) {
            int i = 0;
            for(Attribute attribute : metaFeatures.getAttributes()) {
                values[i][0] = attribute.getName();
                if(attribute.isNominal()) {
                    values[i][1] = example.getNominalValue(attribute);
                } else {
                    values[i][1] = Tools.formatNumber(example.getNumericalValue(attribute));
                }
                i++;
            }
        }
        JTable table = new JTable(values, new String[]{"Feature", "Value"});
        JScrollPane scrollPane = new JScrollPane(table);
        scrollPane.setBorder(BorderFactory.createLineBorder(Color.BLACK));
        JOptionPane.showMessageDialog(parent, scrollPane);
    }
    
    @Override
    public void actionPerformed(ActionEvent event) {
        showFeatureSet();
    }
}
