/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.mlwizard.landmarking;

import java.util.Map;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.set.AttributeSelectionExampleSet;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.learner.tree.DecisionStumpLearner;
import com.rapidminer.operator.learner.tree.NumericalSplitter;
import com.rapidminer.operator.learner.tree.criterions.GainRatioCriterion;
import com.rapidminer.tools.OperatorService;

/**
 * This landmark learner is a {@link DecisionStumpLearner}, where a single
 * decision node is chosen according to the smallest gain ratio.
 * 
 * @author Sarah Daniel
 * 
 */
public class WorstNodeLm extends AbstractLandmarker {
    public WorstNodeLm(ExampleSet exampleSet, Map<String, String> parameters) throws OperatorCreationException, OperatorException {


            this.name = LandmarkingParameters.PARAMETER_WORST_NODE_LM;

            DecisionStumpLearner learner = OperatorService.createOperator(DecisionStumpLearner.class);
            AttributeSelectionExampleSet attributeSelectionExampleSet = getAttributeSelectionExampleset(exampleSet);
            exampleSet = normalizeExampleSet(attributeSelectionExampleSet, parameters);

            learnExampleSet(exampleSet, learner, parameters);
 
    }

    /**
     * @param exampleSet
     *            The {@link ExampleSet} to be modified.
     * @return a new {@link AttributeSelectionExampleSet}, where the selected
     *         attribute has the smallest gain ratio.
     */
    private AttributeSelectionExampleSet getAttributeSelectionExampleset(ExampleSet exampleSet) {
        boolean[] selectedAttributes = new boolean[exampleSet.getAttributes().size()];
        int selectedAttributeIndex = selectAttribute(exampleSet);
        selectedAttributes[selectedAttributeIndex] = true;

        return new AttributeSelectionExampleSet(exampleSet, selectedAttributes);

    }

    /**
     * @param exampleSet
     *            The {@link ExampleSet} used to calculate the gain ratios of
     *            the attributes.
     * @return The index of the attribute with the smallest gain ratio.
     */
    private int selectAttribute(ExampleSet exampleSet) {

        int selectedIndex = 0;

        GainRatioCriterion gainRatioCriterion = new GainRatioCriterion();

        int position = 0;
        double minimalGainSofar = 0;
        for (Attribute attribute : exampleSet.getAttributes()) {
            double gainRatio = 0;

            if (attribute.isNominal()) {
                gainRatio = gainRatioCriterion.getNominalBenefit(exampleSet, attribute);
            }

            else {
                try {
                    gainRatio = gainRatioCriterion.getNumericalBenefit(exampleSet, attribute, new NumericalSplitter(gainRatioCriterion).getBestSplit(exampleSet, attribute));
                } catch (OperatorException e) {
                    e.printStackTrace();
                }
            }

            if (position == 0) {
                minimalGainSofar = gainRatio;
            } else {
                if (gainRatio < minimalGainSofar) {
                    minimalGainSofar = gainRatio;
                    selectedIndex = position;
                }
            }

            position++;
        }

        return selectedIndex;
    }

}
